<p class="logo">
  <a href="{{ pathto(master_doc) }}">
    <img class="logo" src="{{ pathto('_static/equation.png', 1) }}" title="Tartaruga Feliz Equation"/>
  </a>
</p>

<p>
<a href="http://libcrush.org/main/python-crush.git">crush</a> library to control placement in a hierarchy.
</p>


<h3>Useful Links</h3>
<ul>
  <li><a href="http://libcrush.org/main/python-crush/">Code</a></li>
  <li><a href="http://crush.readthedocs.io/">Documentation</a></li>
  <li><a href="http://libcrush.org/main/python-crush/merge_requests">Merge requests</a></li>
  <li><a href="http://libcrush.org/main/python-crush/issues">Issue Tracker</a></li>
  <li><a href="http://libcrush.org/main/python-crush/pipelines">CI</a></li>
</ul>
