{%- extends "basic/layout.html" %}
{%- block extrahead %}
  {{ super() }}
  {% if theme_touch_icon %}
    <link rel="apple-touch-icon" href="{{ pathto('_static/' ~ theme_touch_icon, 1) }}" />
  {% endif %}
  <meta name="viewport" content="width=device-width, initial-scale=0.9, maximum-scale=0.9">
{% endblock %}
{%- block relbar2 %}{% endblock %}
{%- block footer %}
    <div class="footer">
      &copy; Copyright {{ copyright }}.
    </div>
    <a href="http://libcrush.org/main/python-crush.git" class="github">
        <img style="position: absolute; top: 0; right: 0; border: 0;" src="{{ pathto('_static/', 1) }}/forkme.png" alt="Fork me"  class="github"/>
    </a>

{%- endblock %}
